#define RP2040_LCD
#include <rp2040.h>
#include "pico_fmeter.h"

/*
Connection of 2 x 16 or 2 x 20 LC-display
lcd_init() has to be called at startup. 
lcd_cmd() and lcd_charout() are basic commands, lcd_line1() and lcd_line2()
display a complete string starting at position 1.

4-bit mode is used for controlling LCD. D0-D3 are left open.
GPIO-pins 6-13 are used as shown:

  LCD-Pin   Function        RP2040 GPIO

    1         GND            GND-Pin
    2         +5V       +5V supply voltage
    3       Contrast   PWM_3A GPIO6
    4         RS              GPIO7
    5         R/W             GPIO8
    6         E               GPIO9
    7       D0(nc)
    8       D1(nc)
    9       D2(nc)
    10      D3(nc)
    11        D4              GPIO10
    12        D5              GPIO11
    13        D6              GPIO12
    14        D7              GPIO13

LCD-contrast is a little bit different to normal usage. A PWM-signal of 30% of
+3.3 V (eff. 1 V) is generated to drive the low-impedance contrast-input (0.5 mA input current).
If this does not work sufficient you have to use LC-filtering to get DC-voltage
and maybe a cemet-pot to adjust best contrast. Otherwise PWM duty-cycle can be
adjusted.

PWM-frequency is 100 kHz so alternatively a charge-pump can generate a negative
voltage (-Vee) if required for special displays.


2022-05-15
Michael Nowak
http://www.mino-elektronik.de
All information is subject to change!
*/

extern uint32_t SystemCoreClock;

#define LCD_CONTRAST    (1 << 6)
#define LCD_CMD         (1 << 7)
#define LCD_READ        (1 << 8)
#define LCD_STROBE      (1 << 9)
#define LCD_DATA_OFFSET    10

#define MIN_LCD_WIDTH     16
#define DEF_LCD_WIDTH     16
#define MAX_LCD_WIDTH     20
#define DEF_LCD_CONTRAST  30
#define MAX_LCD_CONTRAST  50


#define LCD_INIT_DELAY    50000     // 50 ms
#define LCD_NIBBLE_DELAY  10000     // 10 ms

#define LINE1      0x7f             // add +1 for 1. row
#define LINE2      0xbf             // dto.
#define BUSY_FLAG  0x80             // bit7 
  
uint8_t line_tab[] = {0,LINE1,LINE2,LINE1,LINE2}; // all positions minus 1
uint8_t lcd_line_start[5];

uint8_t lcd_width = DEF_LCD_WIDTH;
uint8_t lcd_contrast = DEF_LCD_CONTRAST;
uint8_t lcd_busy;
uint8_t lcd_exists = 1;             // default

void lcd_impuls(); 

static void delay_us(uint32_t us)
{
  RESETS_CLR->RESET = RESETS_RESET_pwm_Msk;
  while (!(RESETS->RESET_DONE & RESETS_RESET_pwm_Msk));
  PWM->CH1_TOP = 70;                // reduced 1. time delay
  PWM->CH1_CSR =  1;                // enable
  do {
    PWM->INTR =  2;                 // clear CH1-flag
    while(!(PWM->INTR & 2));        // wait until set again
    PWM->CH1_TOP = SystemCoreClock/MHZ_FACTOR;    // for 1 us
  } while(us--);
}   

// init 16 or 20 chars/line
void init_line_width(uint8_t width)
{
int i, offset=1;
  if(width != MIN_LCD_WIDTH) offset += (width-MIN_LCD_WIDTH)/2;       // shift two positions >>
  lcd_width = width;
  line_tab[3] = LINE1 + width;            // for line 3
  line_tab[4] = LINE2 + width;            // for line 4
  for(i=0; i<5; i++)
    lcd_line_start[i] = line_tab[i]+offset;
} 



// PWM_3A for LCD contrast
void set_contrast(uint8_t wert)
{
  IO_BANK0->GPIO6_CTRL = IO_BANK0_GPIO6_CTRL_FUNCSEL_pwm_a_3; // PWM-output
  PWM->CH3_TOP = 99;   // 100%
  PWM->CH3_DIV = (SystemCoreClock/MHZ_FACTOR/10) << PWM_CH0_DIV_INT_Pos;  // 100kHz
  PWM_SET->CH3_CSR = PWM_CH3_CSR_EN_Msk;

  if(wert > MAX_LCD_CONTRAST)                 // max. limit
    wert = DEF_LCD_CONTRAST;
  PWM->CH3_CC = wert;
}  

  
uint8_t lcd_read(void)                        // bit7 is busy-flag, rest: address
{
  uint8_t temp = 0;

  SIO->GPIO_OE_CLR = 0xf << LCD_DATA_OFFSET;  // data lines = input
  
  SIO->GPIO_OUT_SET = LCD_READ;               // set READ
  delay_us(1);
  SIO->GPIO_OUT_SET = LCD_STROBE;             // set STROBE
  delay_us(1);
  temp = (SIO->GPIO_IN >> LCD_DATA_OFFSET) & 0xf;  // first nibble first
  temp <<= 4;
    SIO->GPIO_OUT_CLR = LCD_STROBE;           // first nibble read
  delay_us(1);
    SIO->GPIO_OUT_SET = LCD_STROBE;           // read 2. nibble
  delay_us(1);
  temp |= (SIO->GPIO_IN >> LCD_DATA_OFFSET) & 0xf;
  SIO->GPIO_OUT_CLR = LCD_STROBE;             // done
  SIO->GPIO_OUT_CLR = LCD_READ;               
  delay_us(1);
  
  SIO->GPIO_OE_SET = 0xf << LCD_DATA_OFFSET;  // data lines = output again
  delay_us(1);
  return(temp);                               // busy-flag + address
}

// read status; busy-flag == 0, if ready for new data
uint8_t lcd_status(void)
{
  SIO->GPIO_OUT_CLR = LCD_CMD;                // 0 -> CMD
  delay_us(1);
  return(lcd_read());                         // bit7 is busy-flag, rest: address
} 


void lcd_impuls(void)                         // generate LCD_STROBE
{
  delay_us(1);
  SIO->GPIO_OUT_SET = LCD_STROBE;             // 1 = aktiv
  delay_us(1);
  SIO->GPIO_OUT_CLR = LCD_STROBE;             // 0 = done
  delay_us(1);
}

void lcd_nibble(int8_t c)                     // 4-bit-value to GPIO10 - GPIO13
{
  c >>= 4;
  c &= 0xf;
  SIO->GPIO_OUT_CLR = 0xf << LCD_DATA_OFFSET; // clear data lines
  SIO->GPIO_OUT_SET = c << LCD_DATA_OFFSET;   // set active bits
  lcd_impuls();


}

void lcd_out(int8_t z, int8_t mode)           // send data (mode==1) or CMDs (mode==0)
{
  lcd_busy = 1;
    while(lcd_status() & BUSY_FLAG);          // wait until ready
    if(mode) SIO->GPIO_OUT_SET = LCD_CMD;
    else SIO->GPIO_OUT_CLR = LCD_CMD;
    lcd_nibble(z);                            // upper nibble
    lcd_nibble(z << 4);                       // lower nibble

  lcd_busy = 0;
}

void lcd_cmd(uint8_t z)             // send LCD-CMD
{
  lcd_out(z,0);
}

void lcd_charout(char z)            // send LCD-char
{
  lcd_out(z,1);
}

void lcd_string(char *s)            // send LCD-string
{   
int8_t temp;
  if(lcd_exists) {
    while((temp=*s++) != 0) {
      lcd_charout(temp);
    }
  }
}

void lcd_init_nibble(int8_t c)      // 4-bit for LCD_INIT
{
  lcd_busy = 1;
  lcd_nibble(c);
  lcd_busy = 0;
  delay_us(LCD_NIBBLE_DELAY);       // minimum 10 ms
}

void lcd_line1(char *s)             // LCD-string to line1
{
  if(lcd_exists) {
    lcd_cmd(lcd_line_start[1]);     // 1. line, 1. row
    lcd_string(s);
  }
}
void lcd_line2(char *s)             // LCD-string to line2
{
  if(lcd_exists) {
    lcd_cmd(lcd_line_start[2]);     // 2. line, 1. row
    lcd_string(s);
  }
}

void lcd_cursor_on()
{
  lcd_cmd(0xf);
}

void lcd_cursor_off()
{
  lcd_cmd(0xc);
}

void lcd_line_row(uint8_t line, uint8_t row)  // set position
{
  if(line && row) lcd_cmd(line_tab[line] + row);
}

void clear_lcd_line(uint8_t line, uint8_t n)
{
uint8_t i;  
  if(lcd_exists) {
    lcd_cmd(line_tab[line] + 1);           // 1. position
    for(i=0; i<n; i++) lcd_charout(' ');    // writing ' ' chars
  }  
}  

void lcd_clear()
{
  lcd_cursor_off();
  lcd_cmd(1);
}

void clear_lcd_line1()
{
  clear_lcd_line(1, lcd_width);
}
void clear_lcd_line2()
{
  clear_lcd_line(2, lcd_width);
}
void clear_lcd_line12()
{
  clear_lcd_line(1, lcd_width);
  clear_lcd_line(2, lcd_width);
}

void lcd_result_line1(char *s)            // line1: 16/20 chars
{
  if(lcd_exists) {
    if(lcd_width == DEF_LCD_WIDTH)
      lcd_cmd(lcd_line_start[1]);         // 1. line, 1. row
    else
      lcd_cmd(lcd_line_start[1]-2);       // 1. line, 1. row
    lcd_string(s);
  }
}

void lcd_result_line2(char *s)            // line2: 16/20 chars
{
  if(lcd_exists) {
    if(lcd_width == DEF_LCD_WIDTH)
      lcd_cmd(lcd_line_start[2]);         // 2. line, 1. row
    else
      lcd_cmd(lcd_line_start[2]-2);       // . line, 1. row
    lcd_string(s);
  }
}
 
uint8_t lcd_init(void)                    // connected as shown above
{
// enable GPIOs
  RESETS_CLR->RESET = RESETS_RESET_io_bank0_Msk | 
                      RESETS_RESET_pads_bank0_Msk;
  while((RESETS->RESET_DONE & (RESETS_RESET_io_bank0_Msk | RESETS_RESET_pads_bank0_Msk)) !=
        (RESETS_RESET_io_bank0_Msk | RESETS_RESET_pads_bank0_Msk));

  IO_BANK0->GPIO6_CTRL = IO_BANK0_GPIO6_CTRL_FUNCSEL_pwm_a_3;   // PWM for LCD_CONTRAST 
  IO_BANK0->GPIO7_CTRL = IO_BANK0_GPIO0_CTRL_FUNCSEL_sio_0;
  IO_BANK0->GPIO8_CTRL = IO_BANK0_GPIO0_CTRL_FUNCSEL_sio_0;
  IO_BANK0->GPIO9_CTRL = IO_BANK0_GPIO0_CTRL_FUNCSEL_sio_0;
  IO_BANK0->GPIO10_CTRL = IO_BANK0_GPIO0_CTRL_FUNCSEL_sio_0;
  IO_BANK0->GPIO11_CTRL = IO_BANK0_GPIO0_CTRL_FUNCSEL_sio_0;
  IO_BANK0->GPIO12_CTRL = IO_BANK0_GPIO0_CTRL_FUNCSEL_sio_0;
  IO_BANK0->GPIO13_CTRL = IO_BANK0_GPIO0_CTRL_FUNCSEL_sio_0;
//  SIO->GPIO_OE_SET = 0x3fc0;
  
// Ausgnge push-pull output
  SIO->GPIO_OE_SET = LCD_CMD | LCD_READ | LCD_STROBE; // control lines = output
  SIO->GPIO_OE_SET = 0xf << LCD_DATA_OFFSET;          // data lines = output 

  init_line_width(lcd_width);

  delay_us(LCD_INIT_DELAY);   // 50 ms after reset

  lcd_init_nibble(0x30);      // start with 8 bit mode
  lcd_init_nibble(0x30);      // repeat
  lcd_init_nibble(0x30);      // repeat
  lcd_init_nibble(0x20);      // set 4 bit mode
  lcd_cmd(0x28);              // for 2 or 4 lines
  lcd_cmd(0x28);              // repeat

  lcd_cmd(0x8);               // LCD off            
  lcd_cmd(0x6);               // LCD auto. increment
  lcd_cmd(0x17);              // LCD char. mode; PWR on
  lcd_cmd(0x1);               // LCD clear
  lcd_cmd(0x2);               // LCD home
  
  lcd_cmd(0xc);               // LCD on; no cursor, not blinking
  lcd_cmd(1);                 // LCD clear
  delay_us(LCD_INIT_DELAY);   // 50 ms after init
  clear_lcd_line12();
  set_contrast(lcd_contrast);
  
  return(lcd_status());
}

