/*
reciprocal frequency counter using RP2040

The input signal at FIN1_PIN is sampled at SystemCoreClock/4. Each cycle is
counted by PIO0-SM0-X-register. A falling edge transfers X-register to RXF0-fifo.
DMA-chanal0 is triggerd and writes the new value to 'F1_time_stamp'. Each event
is transferred so DMA_TRANS_COUNT reflects the number of input edges.

After the minimum test intervall 'F1_test_interval' is over the last time-stamp
'F1_time_stamp' and the number of events (DMA_TRANS_COUNT) are captured synchronously.
Calculating the difference of old and new values the effective measurement time 
and events will give a new result: events/time * sample rate.

Sample rate is given by SystemCoreClock/4 which is typically 37.5 MHz.
Default value for 'F1_test_intervall' is 3 tests/s. So the resolution of each measurement
is <= 1*10^-7 (3 / 3.75*10^7) which means 7 digts are available.
Absolute accuracy depends on SystemCoreClock which is PLL-generated based on
local 12 MHz crystal or external TCXO.
For finer adjustment of reference-clock there is 'f_ref_offset'. If connected
an external trimm-pot at ADC0-input allows +/-20 ppm correction. Unconnected
'f_ref_offset' = 0 and external frequency (12 MHZ, 10 MHz) has to be adjusted.


For further explanations please read comments.


DMA-Trigger changed: CH0 and CH10 trigger each other, no ISR needed
2023-03-04
Michael Nowak
http://www.mino-elektronik.de
All information is subject to change!
*/

#include <rp2040.h>
#include <string.h>
#include <stdio.h>
#include "pico_fmeter.h"

#define DMA_COUNT         0x80000000
#define DMA_COUNT_MASK    (DMA_COUNT-1)

#define DEF_DIGITS        7
#define DEF_FACTOR        1
#define DEF_RPM_DIVISOR   1

#define TEMP_STR_LEN      50          // max. strlen: temp
#define MAX_RESULT_STRLEN 20          // max. strlen: result

extern double f_ref_offset;           // +/- 20 ppm vom ADC
extern uint32_t SystemCoreClock;
extern uint8_t lcd_width;             // typ. 16 optional 20

extern void  sende_str0(char *s);
extern void  lcd_line1(char *s);
extern void  lcd_line2(char *s);
extern void  clear_lcd_line12();
extern void  lcd_result_line1(char *s);
extern void  lcd_result_line2(char *s);

double rundung[] = {5e0, 5e-1,5e-2,5e-3,5e-4,5e-5,
                    5e-6,5e-7,5e-8,5e-9,5e-10,5e-11,
                    5e-12,5e-13,5e-14,5e-15};

char f_dim[][10][6]={{" GHz"," MHz"," kHz"," Hz "," mHz"," sec"," ms "," us "," ns "," ps "},
                    {"E+9","E+6","E+3","E+0","E-3","E+0","E-3","E-6","E-9","E-12"}};  // Dimensionen zum Messwert
char u_dim[][5][6]={{" Grpm"," Mrpm"," krpm"," rpm"," mrpm"},
                    {"E+9","E+6","E+3","E+0","E-3"}};

enum dbl_convert {CONVERT_F=0, CONVERT_P, CONVERT_D};
enum show {F_AND_P=0, FREQUENCY, PERIODE, ROTATION};
enum ser_chanal {F1=0, P1, UPM1};     // kind of result

uint32_t  F1_time_stamp,              // of PIO0-SM0
          dma_dummy;

uint32_t  led_count,
          led_on_time = TYP_LED_TIME; // after new meas. ready

int8_t    LCD_locked,                 // suppress LCD-data output
          lcd_function,               // F+P, F only, P only, UPM only or off -1
          update_LCD,                 // LCD new header
          MHz_only,                   // fixed format: x.xxx MHz
          display_format,             // bit0: dimension oder exponent; bit1: . or , 
          F1_prescaler_active,        // not used
          ser_out_chanal;             // ser. data output: F, P, UPM or off -1

char p_str[TEMP_STR_LEN];

double              F1_result;                      // in Hz

int32_t             F1_ref_offset = 0;              // adjustment of F1_reference
uint32_t            F1_interval = DEF_F1_INTERVAL,  // nominal interval
                    F1_timeout = DEF_TIMEOUT,       // missing signal
                    F1_digits = DEF_DIGITS,         // constantly 7
                    F1_factor  = DEF_FACTOR,        // constantly 1
                    F1_rpm_divisor  = DEF_RPM_DIVISOR;  // the same
                    
volatile uint8_t    F1_result_valid;                // valid result
                                   
volatile uint32_t   F1_test_interval,               // in ms
                    F1_timeout_counter;             // in ms
uint32_t            F1_start_time,                  // 1. time of test
                    F1_end_time,                    // last time of test
                    F1_test_time;                   // eff. intervall
                    
uint32_t            F1_start_event,                 // counts at start
                    F1_end_event,                   // counts at end
                    F1_test_events;                 // numer of events

// conditional LCD data output
void lcd_result1(char *s)
{
  if(!LCD_locked) lcd_result_line1(s);
}  

void lcd_result2(char *s)
{
  if(!LCD_locked) lcd_result_line2(s);
} 

// convert double to ASCII-string
void wandel_Fx(char *header, double x, char show_period, char *s, uint8_t digits)    // Anzeige von F, P oder UPM
{
int i,j,dez_punkt,dimension;
char *p;
  if(show_period == CONVERT_F && MHz_only)
  {
    if(x >= 1e6) digits--;
    if(x >= 1e7) digits--;
    if(x >= 1e8) digits--;
    if(x >= 1e9) digits--;
    sprintf(s,"%.*f MHz",digits,x*1e-6);   // fixed dimension "MHz"
    if(*s == '0') *s = ' ';
    return;
  }    

  dez_punkt = 0;
  dimension = 3;
  p = header;
  while(*p) *s++ = *p++;                    // "P = " for period
  if(x >= 0.001) {                          // 1 mHz = lowest limit
    if(show_period == CONVERT_P) {
      x = 1/x;                              // Hz -> s
      dimension=5;                          // startindex
    }
    while(x<1.0) {x*=1000.0;dimension++;}   // scale to 1.0 -> 999.9999
    while(x>=1000.0) {x*=0.001;dimension--;}
    while(x >= 10.0) {
      x *= 0.1;
      dez_punkt++;
    }
    x += rundung[digits];                   // round
    if(x >= 10.0) {			                    // overflow
      x *= 0.1;                             // correction
      dez_punkt++;
      if(dez_punkt > 2) {		                // correct dimension too
        dimension--;
        dez_punkt = 0;
      }
    }
  } else {
    x = 0.0; dez_punkt=0; dimension=3;      // display 0.000000 Hz
  }
  
  for(i=0;i<digits;i++) {
    j = (uint8_t)x;
    *s++ = (j+'0');
    if(i == dez_punkt) {
      if(display_format & 2)                // use '.' or ','
        *s++ = (',');
      else
        *s++ = ('.');
    }  
    x -= j;
    x *= 10;
  }

  if(show_period == CONVERT_D)               // show RPM
    p = u_dim[display_format&1][dimension]; // eet dimension or exponent
  else                                      // show F or P
    p = f_dim[display_format&1][dimension]; // set dimension or exponent
  while(*p) *s++ = *p++;                    // add dimension or exponent
  *s = 0;                                   // finish string
  s[MAX_RESULT_STRLEN] = 0;                 // be sure to terminate
}

void Fx_str_mittig(char *header, double x, char show_period, char *s, uint8_t digits)    // display F, P oder RPM
{
char p[50];
int i,j;
  *p = 0;
  for(i = 0; i < MAX_RESULT_STRLEN; i++)    // clear string
    s[i] = ' ';
  s[MAX_RESULT_STRLEN] = 0;                 // be sure to terminate
  
  wandel_Fx(header,x,show_period,p,digits); // convert to ASCII
  i = strlen(p);
  i = lcd_width - i;                        // center for 20 chars display
  if(i < 0) i = 0;                          // prevent underflow
  j = i/2;
  i = 0;
  while(p[i]) s[j++] = p[i++];              // copy centred string
  s[lcd_width] = 0;                         // be sure to terminate
} 

/*
 PIO0-SM0:
 Count cycles of SystemCoreClock/PIO_TIME_DIV (=4) with X-register.
 Send X-register to RXF0 after falling edge of input signal.
*/
void init_PIO0(void)
{
// activate PIO0   
  RESETS_CLR->RESET = RESETS_RESET_pio0_Msk;
  while (!(RESETS->RESET_DONE & RESETS_RESET_pio0_Msk));

// init FIN1_PIN to SM0 
  IO_BANK0->GPIO2_CTRL = IO_BANK0_GPIO0_CTRL_FUNCSEL_sio_0;  // GPIO-input
  PADS_BANK0->GPIO2 = PADS_BANK0_GPIO0_IE_Msk |              // pullup + schmitttrigger
                      PADS_BANK0_GPIO0_PUE_Msk | 
                      PADS_BANK0_GPIO0_SCHMITT_Msk;
  
  PIO0->SM0_PINCTRL |= 1<<PIO0_SM0_PINCTRL_SIDESET_BASE_Pos | 1<<PIO0_SM0_PINCTRL_SET_BASE_Pos;
  PIO0->SM0_EXECCTRL = (31 << PIO0_SM0_EXECCTRL_WRAP_TOP_Pos) | 
                        (0 << PIO0_SM0_EXECCTRL_WRAP_BOTTOM_Pos) |
                        (FIN1_PIN  << PIO0_SM0_EXECCTRL_JMP_PIN_Pos);  // pin-nr for JMP
// ADR0  
  PIO0->INSTR_MEM0 = 0x0241;      // jmp x-- -> ADR1, 2 x delay
// ADR1
  PIO0->INSTR_MEM1 = 0x00c0;      // jmp FIN1_PIN -> ADR0; wait until 1 -> 0
// input-pin 1 -> 0
  PIO0->INSTR_MEM2 = 0x4020;      // mov x->isr, send X
// ADR3
  PIO0->INSTR_MEM3 = 0x0044;      // jmp x-- -> ADR4
// ADR4  
  PIO0->INSTR_MEM4 = 0x00c1;      // jmp FIN1_PIN; if  0-> 1 restart at ADR0
  PIO0->INSTR_MEM5 = 0x0103;      // jmp ADR3, 1 x delay
  
  PIO0_SET->SM0_SHIFTCTRL = PIO0_SM0_SHIFTCTRL_FJOIN_RX_Msk | // FIFO = 8
                            PIO0_SM0_SHIFTCTRL_AUTOPUSH_Msk;  // autopush
  PIO0_SET->CTRL = (1 << (PIO0_CTRL_SM_ENABLE_Pos));          // start
}


// init DMA_CH0
// CH0 chained to CH10 and CH10 to CH0 again
// no ISR is needed
void init_dma_pio0(void)
{
  RESETS_CLR->RESET = RESETS_RESET_dma_Msk;
  while (!(RESETS->RESET_DONE & RESETS_RESET_dma_Msk));

// CH0 counts events and reads time stamps 
  DMA->CH0_READ_ADDR = (uint32_t) &PIO0->RXF0;
  DMA->CH0_WRITE_ADDR = (uint32_t) &F1_time_stamp;
  DMA->CH0_TRANS_COUNT = DMA_COUNT;
  DMA->CH0_CTRL_TRIG =  4<<DMA_CH0_CTRL_TRIG_TREQ_SEL_Pos |     // SM0-RXF0
                        10 << DMA_CH0_CTRL_TRIG_CHAIN_TO_Pos |  // trigger CH10
                        2 << DMA_CH0_CTRL_TRIG_DATA_SIZE_Pos |  // word
                        DMA_CH0_CTRL_TRIG_EN_Msk;               // enable

// triggered by CH0 and triggers CH0 again  
  DMA->CH10_READ_ADDR = (uint32_t) &dma_dummy;                  // dummy location
  DMA->CH10_WRITE_ADDR = (uint32_t) &dma_dummy;
  DMA->CH10_TRANS_COUNT = 1;                                    // only 1 transfer
  DMA->CH10_CTRL_TRIG =  0x3f<<DMA_CH0_CTRL_TRIG_TREQ_SEL_Pos | // free running
                        0 << DMA_CH0_CTRL_TRIG_CHAIN_TO_Pos |   // Trigger CH0  
                        2 << DMA_CH0_CTRL_TRIG_DATA_SIZE_Pos |  // word
                        DMA_CH0_CTRL_TRIG_EN_Msk;               // enable
}

// 1 kHz 
void PWM_IRQ_WRAP_IRQHandler(void)
{
static uint32_t F1_time_stamp_old;
  if(PWM->INTS & 1) {                           // PWM_CH0?
    PWM->INTR = 1;                              // clear flag
    F1_test_interval++;
    F1_timeout_counter++;
    if(F1_time_stamp != F1_time_stamp_old) {    // new count?
      F1_time_stamp_old = F1_time_stamp;        // yes
      F1_timeout_counter = 0;                   // reset timeout
    }
    if(led_count) {                             // ready-LED
      SIO->GPIO_OUT_SET = (1 << LOCAL_LED);
      led_count--;
      if(!led_count)
        SIO->GPIO_OUT_CLR = (1 << LOCAL_LED);
    }  
  }
  __DSB();
}

// ms-timing for measurement
void init_F1_timer(void)
{
  RESETS_CLR->RESET = RESETS_RESET_pwm_Msk;
  while (!(RESETS->RESET_DONE & RESETS_RESET_pwm_Msk));
  IO_BANK0->GPIO25_CTRL = IO_BANK0_GPIO0_CTRL_FUNCSEL_sio_0;  // output for LED
  SIO->GPIO_OE_SET = 1 << LOCAL_LED; 
  
  PWM->CH0_DIV = 100 << PWM_CH0_DIV_INT_Pos;      // set prescaler to 100
  PWM->CH0_TOP = SystemCoreClock/100/1000-1;    	// 1 kHz
  NVIC_EnableIRQ(PWM_IRQ_WRAP_IRQn);
  NVIC_SetPriority(PWM_IRQ_WRAP_IRQn,3);
  
  PWM->CH0_CSR =  1;                              // enable ms-timer
  PWM->INTE =  1;                                 // enable ISR
}

// actual time + events
void get_time_count(uint32_t *time, uint32_t *count)
{
uint32_t temp1, temp2;      // temp registers for faster reading
  DMA_CLR->CH0_CTRL_TRIG = DMA_CH0_CTRL_TRIG_EN_Msk;  // stop DMA
  temp1 =   DMA->CH0_AL1_TRANS_COUNT_TRIG;            // count
  temp2 = F1_time_stamp;                              // get time stamp
  DMA_SET->CH0_CTRL_TRIG = DMA_CH0_CTRL_TRIG_EN_Msk;  // continue DMA
  *count = temp1;
  *time  = temp2;
}  
     
// F1_messung is called continously and updated LCD + ser. output if new result
void F1_messung(uint8_t init_timer)
{
static uint8_t  init = 0, old_funtion = 255;
uint8_t eff_digits;
static char p_str[TEMP_STR_LEN];
static char s[TEMP_STR_LEN];
double SM0_clock_temp;

  if(init_timer) { 
    init = 0;                           // init all
    init_F1_timer();
    init_dma_pio0();
    init_PIO0(); 
  }
   
  if(!init) {
    F1_test_interval =  F1_interval;    // finish 1. test at once to get start-values
    F1_result_valid  = 0;               // suppress 1. result
    F1_timeout_counter = 0; 
    init = 1;
    update_LCD = 1;
  }
 
  if(!LCD_locked && (lcd_function != old_funtion || update_LCD)) {  
    update_LCD = 0;
    if(lcd_function > ROTATION) 
      lcd_function = F_AND_P;           // limit to legal value
    old_funtion = lcd_function;
    clear_lcd_line12();
    switch(lcd_function) {
      case F_AND_P:
        lcd_line1("   Signal F1?   ");
        lcd_line2("   Signal P1?   ");
        break;
      case FREQUENCY:
        lcd_line1("  Frequency F1   ");
        lcd_line2("    Signal?     ");
        break;
      case PERIODE:
        lcd_line1("   Period  F1   ");
        lcd_line2("    Signal?     ");
        break;
    case ROTATION:
        lcd_line1("     RPM F1     ");
        lcd_line2("    Signal?     ");
        break;
    }
  }
 
  if(F1_timeout_counter >= F1_timeout) {          // missing input signal
    init = 0;                                     // start init again
  }  
  if(F1_test_interval >= F1_interval) {           // minimum time is over
    get_time_count(&F1_end_time, &F1_end_event);
    F1_test_time = F1_start_time-F1_end_time;     // calc. eff. time
    F1_start_time = F1_end_time;                
    F1_test_events = (F1_start_event-F1_end_event) & DMA_COUNT_MASK; // calc. eff. events
    F1_start_event = F1_end_event;
    if(F1_test_events) {                          // there must be >= 1 event!
// SM0_clock_temp = corrected F-ref
      SM0_clock_temp = (SystemCoreClock+f_ref_offset)/PIO_TIME_DIV;
      F1_result = (double)F1_test_events/(double)F1_test_time * SM0_clock_temp;
      F1_test_interval = 0;                       // start new measurement

      if(F1_result_valid) {                       // if not locked
        eff_digits = F1_digits;
        if(F1_prescaler_active)
          F1_result *= (double)F1_factor;         // for prescaler if needed

        led_count = led_on_time;
// send to UART0        
        if(ser_out_chanal == P1) {
          wandel_Fx("",F1_result, CONVERT_P, s, eff_digits);
          sende_str0(s);
        }
        if(ser_out_chanal == UPM1) {
          wandel_Fx("",F1_result*60.0/F1_rpm_divisor, CONVERT_D, s, eff_digits);
          sende_str0(s);
        }
        wandel_Fx("",F1_result, CONVERT_F, s, eff_digits);
        if(ser_out_chanal == F1) 
          sende_str0(s);
// send to LCD        
        switch(lcd_function) {
          case F_AND_P:
            Fx_str_mittig("F = ",F1_result, CONVERT_F, p_str, eff_digits);
            lcd_result1(p_str);
            Fx_str_mittig("P = ",F1_result, CONVERT_P, p_str, eff_digits);
            lcd_result2(p_str);
            break;
          case FREQUENCY:
            Fx_str_mittig("",F1_result, CONVERT_F, p_str, eff_digits);
            lcd_result2(p_str);
            break;
          case PERIODE:
            Fx_str_mittig("",F1_result, CONVERT_P, p_str, eff_digits);
            lcd_result2(p_str);
            break;
          case ROTATION:
            Fx_str_mittig("",F1_result*60.0/F1_rpm_divisor, CONVERT_D, s, eff_digits);
            lcd_result2(s);
            break;
        }
      } else F1_result_valid  = 1;                // show next result
    }
  }  
}  